local ffi = require("ffi")

-- Import math functions for convenience
local min, max = math.min, math.max
local sin, cos, tan = math.sin, math.cos, math.tan
local pow, rad = math.pow, math.rad
local floor, ceil = math.floor, math.ceil
local log = math.log
local bor, band, lshift, rshift, bxor, bnot = bit.bor, bit.band, bit.lshift, bit.rshift, bit.bxor, bit.bnot

-- Set default filter for graphics
love.graphics.setDefaultFilter('nearest', 'nearest')

-- Clamp function to restrict a value between a minimum and maximum
local function clamp(d, min, max)
  local t = d < min and min or d
  return t > max and max or t
end

-- Camera configuration
local cam = {
  x = 0,
  y = 0,
  height = 0,
  speed = 100,
  angle = 45,
  pitch = rad(20),
  fov = tan(rad(90) / 2),
  zNear = 128,
  zFar = 1024,
  offset = -100
}

-- Frame buffer configuration
local frameBuffer = {
  width = GAME_WIDTH,
  height = GAME_HEIGHT,
  scale = SCALE_X,
  data = nil,
  img = nil,
  ptr = nil
}

-- Initialize frame buffer
frameBuffer.data = love.image.newImageData(frameBuffer.width, frameBuffer.height, "rgba8")
frameBuffer.ptr = ffi.cast('uint32_t*', frameBuffer.data:getFFIPointer())
frameBuffer.img = love.graphics.newImage(frameBuffer.data)

-- Data buffers
local data = {
  yBuffer = love.data.newByteData(frameBuffer.width * ffi.sizeof("uint16_t")),
  depthLUT = love.data.newByteData(cam.zFar * ffi.sizeof("double"))
}

local ptr = {
  yBuffer = ffi.cast('uint16_t*', data.yBuffer:getFFIPointer()),
  depthLUT = ffi.cast('double*', data.depthLUT:getFFIPointer())
}

-- Terrain textures and height maps
local terrainTex = {
  love.image.newImageData("color_1.png"),
  love.image.newImageData("color_2.png"),
  love.image.newImageData("color_3.png"),
  love.image.newImageData("color_4.png")
}

local terrainHeight = {
  love.filesystem.newFileData("height_1.r32"),
  love.filesystem.newFileData("height_2.r32"),
  love.filesystem.newFileData("height_3.r32"),
  love.filesystem.newFileData("height_4.r32")
}

local terrainPtr = {
  ffi.cast('uint32_t*', terrainTex[1]:getFFIPointer()),
  ffi.cast('uint32_t*', terrainTex[2]:getFFIPointer()),
  ffi.cast('uint32_t*', terrainTex[3]:getFFIPointer()),
  ffi.cast('uint32_t*', terrainTex[4]:getFFIPointer())
}

local terrainHeightPtr = {
  ffi.cast('float*', terrainHeight[1]:getFFIPointer()),
  ffi.cast('float*', terrainHeight[2]:getFFIPointer()),
  ffi.cast('float*', terrainHeight[3]:getFFIPointer()),
  ffi.cast('float*', terrainHeight[4]:getFFIPointer())
}

local tileSize = terrainTex[1]:getWidth()
local tileX, tileY, tileIndex = 0, 0, 0
local terrainSize, tileSizeBits = tileSize * 2, log(tileSize, 2)

-- Function to draw a vertical line on the frame buffer
local function drawLine(x, y1, y2, color, frameBuffer)
  y1 = clamp(y1, 0, frameBuffer.height - 1)
  y2 = clamp(y2, 0, frameBuffer.height - 1)
  for y = y1, y2 do
    frameBuffer.ptr[y * frameBuffer.width + x] = color
  end
end

-- Function to fill the depth lookup table
local function fillDepthLUT(arr, cam, exponent)
  local distance = cam.zFar
  for i = 0, distance do
    arr[i] = pow(i / distance, exponent) * distance * 2
  end
end
fillDepthLUT(ptr.depthLUT, cam, 3)

-- Function to get coordinates within a tile
local function getCoords(x, y, tileSize)
  return lshift(band(y - 0.5, tileSize - 1), tileSizeBits) + band(x - 0.5, tileSize - 1)
end

-- Function to get the index of a tile
local function getTileIndex(x, y)
  tileX = rshift(x, tileSizeBits)
  tileY = rshift(y, tileSizeBits)
  tileIndex = lshift(tileY, log(rshift(terrainSize, tileSizeBits), 2)) + tileX
  return tileIndex + 1
end

-- Function to clear the frame buffer
local function clearFrameBuffer(frameBuffer, color)
  for i = 0, frameBuffer.width * frameBuffer.height -1 do
    frameBuffer.ptr[i] = color
  end
end

-- Main render function
local function Render(frameBuffer, cam, terrainHeightPtr, yBuffer, depthLUT, terrainPtr, cloudsText2)
  clearFrameBuffer(frameBuffer, 0xFFFF0000)

  local scalingRatio = 0.25
  local heightScale = tileSize * scalingRatio

  local sinphi = sin(rad(cam.angle))
  local cosphi = cos(rad(cam.angle))

  local K0 = frameBuffer.width / 2
  local K1 = (frameBuffer.height / 2) + K0 * tan(-cam.pitch)
  local K2 = (heightScale / cam.fov) * K0

  for i = 0, frameBuffer.width - 1 do
    yBuffer[i] = frameBuffer.height - 1
  end

  local csw, snw = cam.fov * cosphi, cam.fov * sinphi
  local xs, ys = cosphi - snw, sinphi + csw
  local csw2, snw2 = csw / frameBuffer.width * 2, snw / frameBuffer.width * 2
  local xMin, yMin = 0, 0
  local xMax, yMax = terrainSize - 1, terrainSize - 1

  for j = cam.zNear, cam.zFar do
    local depth = depthLUT[j]
    local coordX = cam.x + depth * xs
    local coordY = cam.y + depth * ys
    local K3 = K2 / depth
    local K4 = K1 - K3 * cam.height / heightScale
    local deltaX = depth * snw2
    local deltaY = -depth * csw2

    for i = 0, frameBuffer.width - 1 do
      if coordX < xMin or coordX > xMax or coordY < yMin or coordY > yMax then
        goto continue
      else
        local tileIndex = getTileIndex(coordX, coordY)
        local coordsUV = getCoords(coordX, coordY, tileSize)
        local TerrainHeight = 1.0 - terrainHeightPtr[tileIndex][coordsUV]
        local columnHeight = TerrainHeight * K3 + K4

        if columnHeight < yBuffer[i] and yBuffer[i] >= 0 and yBuffer[i] < frameBuffer.height then
          drawLine(i, ceil(columnHeight), yBuffer[i], terrainPtr[tileIndex][coordsUV], frameBuffer)
          yBuffer[i] = columnHeight
        end
      end

      ::continue::
      coordX = coordX + deltaX
      coordY = coordY + deltaY
    end
  end
  frameBuffer.img:replacePixels(frameBuffer.data)
end

-- Update function called every frame
function love.update(dt)
  local deltaX, deltaY = 0, 0

  -- Get input direction
  if love.keyboard.isDown("w") then
    deltaX = 1
  elseif love.keyboard.isDown("s") then
    deltaX = -1
  end

  if love.keyboard.isDown("a") then
    deltaY = -1
  elseif love.keyboard.isDown("d") then
    deltaY = 1
  end

  if love.keyboard.isDown("q") then
    cam.angle = cam.angle + (90 * dt)
  elseif love.keyboard.isDown("e") then
    cam.angle = cam.angle - (90 * dt)
  end

  local rotatedDx = deltaX * cos(rad(cam.angle)) + deltaY * sin(rad(cam.angle))
  local rotatedDy = deltaX * sin(rad(cam.angle)) - deltaY * cos(rad(cam.angle))
  cam.x = cam.x + rotatedDx * cam.speed * dt
  cam.y = cam.y + rotatedDy * cam.speed * dt

  Render(frameBuffer, cam, terrainHeightPtr, ptr.yBuffer, ptr.depthLUT, terrainPtr, ptr.cloudsText2)
end

-- Draw function called every frame
function love.draw()
  love.graphics.draw(frameBuffer.img, 0, 0, 0, frameBuffer.scale)
  love.graphics.print(love.timer.getFPS() .. ' FPS', 10 * frameBuffer.scale, 20 * frameBuffer.scale, 0, frameBuffer.scale)
  love.graphics.print(tileX .. " | " .. tileY .. " | " .. tileIndex, 10 * frameBuffer.scale, 40 * frameBuffer.scale, 0, frameBuffer.scale)
  love.graphics.print(cam.x .. " | " .. cam.y .. " | " .. cam.height, 10 * frameBuffer.scale, 60 * frameBuffer.scale, 0, frameBuffer.scale)

-- COMMENT THIS OUT AND IT RUNS FOREVER WITHOUT CRASHING
  love.graphics.print(collectgarbage("count"), 10 * frameBuffer.scale, 80 * frameBuffer.scale, 0, frameBuffer.scale)
-- COMMENT THIS OUT AND IT RUNS FOREVER WITHOUT CRASHING
end


function love.keypressed(key)
  if key == 'escape' then love.event.quit() end
end

local TICK_RATE = 1 / 60
local MAX_FRAME_SKIP = 15

-- Main loop function
function love.run()
  if love.load then love.load(love.arg.parseGameArguments(arg), arg) end
  if love.timer then love.timer.step() end

  local lag = 0.0

  return function()
    -- Process events
    if love.event then
      love.event.pump()
      for name, a, b, c, d, e, f in love.event.poll() do
        if name == "quit" then
          if not love.quit or not love.quit() then
            return a or 0
          end
        end
        love.handlers[name](a, b, c, d, e, f)
      end
    end

    -- Cap the number of frames that can be skipped to prevent lag accumulation
    if love.timer then lag = min(lag + love.timer.step(), TICK_RATE * MAX_FRAME_SKIP) end

    while lag >= TICK_RATE do
      if love.update then love.update(TICK_RATE) end
      lag = lag - TICK_RATE
    end

    if love.graphics and love.graphics.isActive() then
      if love.draw then love.draw() end
      love.graphics.present()
    end
  end
end
